"""RTMP client connection."""

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
from warnings import warn

import handshake
import packet
import select
import socket
import struct
import threading

SOCKET_TIMEOUT = 60                         # Default timeout in seconds

class RTMPSocket(socket.socket):
    def __init__(self, type):
        socket.socket.__init__(self, socket.AF_INET, type)
        self.timeout = SOCKET_TIMEOUT
        self.bytes_read = 0

    def read(self, bytes):
        """Read a specific number of bytes from the socket."""
        result = ""
        while len(result) != bytes:
            ready = select.select([self], [], [], self.timeout)[0]
            if len(ready) == 0:
                raise RTMPError('Reading RTMPSocket timed out.')
            result += self.recv(bytes)
        self.bytes_read += bytes
        return result

    def recv(self, *args):
        """recv(buffersize[, flags]) -> data"""
        result = socket.socket.recv(self, *args)
        self.bytes_read += len(result)
        return result


class RTMPConnection:
    """A connection using the Real Time Messaging Protocol."""
    def __init__(self, host, port=1935):
        """Create a new connection object (in an unconnected state)."""
        self.address = (host, port)         # Connect to this address.
        self.socket = None                  # RTMPSocket object.
        self.last_call_id = {}              # Channel # -> Last call() ID.
        self.last_stream_id = {}            # Channel # -> Last stream ID.
        self.read_thread = None             # Socket reading thread.
        self.audio_buffer = StringIO()      # Audio stream.
        self.audio_lock   = threading.Lock()
        self.video_buffer = StringIO()      # Video stream.
        self.video_lock   = threading.Lock()

        # calls() wait for results in waiting_calls.  This should not be
        # modified without acquiring the waiting_calls_lock.
        self.waiting_calls = {}
        self.waiting_calls_lock = threading.Lock()

        self.last_remote_call_id = {}       # Channel # -> last call() ID.
        self.last_remote_call_id_lock = threading.Lock()

        # Client-side functions that the server may call.  The return values
        # will be passed back to the server.
        self.client_functions = {}

        # onMetaData is a special packet, so we have a separate method to set
        # its callback.
        self.metadata_function = None

    def __del__(self):
        self.close()

    def connect(self):
        """Connect to the host and port specified in __init__.
                
        A NetStream object can optionally be used, specified as a dict.  Some
        useful keys in this dict are app, flashVer, swfUrl, tcUrl, fpad,
        audioCodecs, videoCodecs, pageUrl.

        """
        if self.socket != None:
            self.socket.close()
        self.socket = RTMPSocket(socket.SOCK_STREAM)
        self.socket.connect(self.address)
        handshake.perform(self.socket)
        self.read_thread = threading.Thread(target=self._reader)
        self.read_thread.start()

    def make_header(self, channel, type, size, time=0, stream=0):
        """Return an RTMP header for a new AMF packet."""
        if channel in self.last_stream_id and \
           self.last_stream_id[channel] == stream:
            header_size = 0x01
        else:
            self.last_stream_id[channel] = stream
            header_size = 0x00

        firstByte = (header_size << 6) | (channel & 0x3F)
        result = ''.join((chr(firstByte),\
                          struct.pack('>I', time)[1:4],\
                          struct.pack('>I', size)[1:4],\
                          chr(type)))

        if header_size == 0x00: # Note: The stream is LITTLE ENDIAN.
            result += struct.pack('<I', stream)

        return result


    def add_client_function(self, name, function):
        """Add a client-side function to this connection.

        When the server calls the function named 'name', the function will
        be evaluated and all of its return values will be passed back to the
        server.  If the function returns None instead of a list or tuple, then
        no message will be passed back to the server.

        For example, I've seen "onBWCheck" return (None, 3).

        """
        self.client_functions[name] = function

    def set_metadata_function(self, function):
        """Add a callback to receive stream metadata.
        
        The function should take one argument, a dict of metadata.
        
        """
        self.metadata_function = function

    def remove_client_function(self, f):
        """Remove a client-side function from this connection."""
        del self.client_functions[f]


    def call(self, function, *args, **options):
        """Perform a remote function call.

        The following named arguments are recognized:

        - 'channel' will use a specific channel number.  By default, channel=3.
        - 'callback' will call the specified python function with the result as
          arguments.  If no callback is given, call() will block until the
          result is received.
        - 'no_retval' means that you do not expect a return value, so this
          call() will return immediately.
        - 'result_id' will send a hard-coded function call id to the server,
          and no response will be expected.  This should not be used under
          normal circumstances.
        - 'timestamp' will set the RTMP timestamp for this call.
        - 'stream' will set the stream ID for this call.  You can get a stream
          ID by the remote call createStream().

        """
        if 'channel' in options: channel = options['channel']
        else: channel = 3
        if 'callback' in options: callback = options['callback']
        else: callback = None
        if 'no_retval' in options: no_retval = options['no_retval']
        else: no_retval = False
        if 'result_id' in options: result_id = options['result_id']
        else: result_id = None
        if 'timestamp' in options: timestamp = options['timestamp']
        else: timestamp = 0
        if 'stream' in options: stream = options['stream']
        else: stream = 0

        # Make an ID for this call so we can get the result later.
        if result_id != None:
            id = result_id
        else:
            if not channel in self.last_call_id:
                self.last_call_id[channel] = -1
            if channel in self.last_remote_call_id:
                my_last_id   = self.last_remote_call_id[channel]
                your_last_id = self.last_call_id[channel]
                id = max(my_last_id, your_last_id) + 1
            else:
                id = self.last_call_id[channel] + 1
            self.last_call_id[channel] = id
        # Serialize the function call using AMF.
        body = packet.serialize(function, id, *args)
        type = packet.TYPE_FUNCTION
        output = self.make_header(channel, type, len(body), timestamp, stream)
        chunks = chunk_message(body)
        separator = chr(0xc0 | channel)
        output += separator.join(chunks)
        self.socket.send(output)
        if result_id != None or no_retval == True:
            return                          # No return value expected.

        if callback != None:
            self._set_waiting_call(id, callback)
            results = None
        else:
            results = self._wait_for_call_result(id)
        return results

    def set_buffer_time(self, time):
        """Set the buffer time, probably in milliseconds.
        
        In Red5, this can be between 1000 and 5000.

        """
        channel = 2
        value2 = 1 # In Red5, it seems to be non-zero to set the time.
        data = struct.pack('>HII', 0x03, value2, time)
        header = self.make_header(channel, packet.TYPE_PING, len(data))
        self.socket.send(header + data)

    def _reader(self):
        """Reading thread.  One thread per connection."""
        parser = packet.PacketReader()
        while self.socket != None:
            try:
                ready = select.select([self.socket], [], [], 1.0)[0]
            except:
                break
            if len(ready) != 0:
                data = self.socket.recv(16384)
                results = parser.parse(data)
                for p in results:
                    if p.type == packet.TYPE_AUDIO:
                        self._add_new_buffer(p.data, self.audio_buffer, \
                                             self.audio_lock)
                    elif p.type == packet.TYPE_VIDEO:
                        self._add_new_buffer(p.data, self.video_buffer, \
                                             self.video_lock)
                    elif p.type == packet.TYPE_FUNCTION:
                        self._call_callback(p)
                    elif p.type == packet.TYPE_METADATA:
                        self._handle_metadata(p)
                    else:
                        warn("Ignoring "+str(p)+" "+repr(p.data))
                # Update remote call id's
                for num, channel in parser.channels.items():
                    self.last_remote_call_id[num] = channel.call_id
        # end while

    def close(self):
        """Close this connection."""
        if self.socket != None:
            self.socket.close()
            self.socket = None
            self.read_thread.join()
        self.waiting_calls_lock.acquire()
        try:
            for function, value in self.waiting_calls.items():
                if isinstance(self.waiting_calls[function], threading._Event):
                    # Wake up function calls that are still waiting for results.
                    self.waiting_calls[function].set()
                    self.waiting_calls[function] = None
        finally:
            self.waiting_calls_lock.release()


    def _add_new_buffer(self, data, buffer, lock):
        """Add this a/v data to the buffer."""
        lock.acquire()
        try:
            buffer.write(data)
        finally:
            lock.release()

    def _read_and_clear_buffer(self, buffer, lock):
        """Read from and clear a lockable a/v buffer."""
        lock.acquire()
        try:
            result = buffer.getvalue()
            buffer.seek(0)
            buffer.truncate()
        finally:
            lock.release()
        return result

    def read_video(self):
        """Flush the incoming video buffer, returning the contents.

        The result may be zero-length if there is nothing in the buffer."""
        return self._read_and_clear_buffer(self.video_buffer, self.video_lock)

    def read_audio(self):
        """Flush the incoming audio buffer, returning the contents.

        The result may be zero-length if there is nothing in the buffer."""
        return self._read_and_clear_buffer(self.audio_buffer, self.audio_lock)

    def send_bytes_read(self):
        """Ping the server with the number of bytes read so far."""
        header = self.make_header(channel=2, type=packet.TYPE_BYTES_READ, \
                                  size=4)
        body = struct.pack('>I', (self.socket.bytes_read))
        self.socket.send(header + body)
        # XXX self.socket.bytes_read is wrong.  recv() isn't overridden.


    def _call_callback(self, p):
        """Call the callback for this packet, or wake up a blocking call()."""
        id   = int(p.id)
        args = p.args
        if args[0] == None: 
            args = args[1:]     # Looks like we can safely ignore this.
        if p.function == "_result" or p.function == "_error":
            # This is the result of a function we called earlier.
            if len(args) == 0:
                args = None         # No return value.
            elif len(args) == 1:
                args = args[0]      # Single return value.
            self.waiting_calls_lock.acquire()
            try:
                if not id in self.waiting_calls:
                    warn("Got a response for "+str(id)+", a function we know"\
                         "nothing about.")
                else:
                    if isinstance(self.waiting_calls[id], threading._Event):
                        # Let the event finish even if it's an error, or call()
                        # will hang.
                        event = self.waiting_calls[id]
                        self.waiting_calls[id] = (p.function, args)
                        event.set()
                    elif p.function != "_error":
                        # Use the callback if there's a result.
                        self.waiting_calls[id](*args)
                        del self.waiting_calls[id]
            finally:
                self.waiting_calls_lock.release()
            if p.function == "_error":
                raise RTMPFunctionError, str(p.args)
        else:
            # It's not a return value; the server is asking us to do something!
            if p.function in self.client_functions:
                results = self.client_functions[p.function](*p.args)
                if results != None:
                    self.call('_result', result_id=p.id, *results)
            else:
                warn("No client function for "+str(p.function))

    def _set_waiting_call(self, id, callback):
        """Add a new waiting call to the list.  (Synchronized method.)"""
        self.waiting_calls_lock.acquire()
        try:
            self.waiting_calls[id] = callback
        finally:
            self.waiting_calls_lock.release()

    def _wait_for_call_result(self, id):
        """Wait for a call() to return.  (Synchronized method.)"""
        event = threading.Event()
        self.waiting_calls_lock.acquire()
        try:
            self.waiting_calls[id] = event
        finally:
            self.waiting_calls_lock.release()
        event.wait()

        self.waiting_calls_lock.acquire()
        try:
            # The event has been replaced with the call's results.
            if self.waiting_calls[id] != None:
                success, results = self.waiting_calls[id]
            else:
                return False # the connection is being deleted
            del self.waiting_calls[id]
        finally:
            self.waiting_calls_lock.release()

        if success == "_error":
            raise RTMPFunctionError, str(results)
        else:
            return results

    def _handle_metadata(self, p):
        """Handle a metadata packet."""
        if len(p.data) != 2:
            warn("Ignoring Metadata packet with %d fields." % len(p.data))
            return
        if str(p.data[0]) != 'onMetaData' or self.metadata_function == None:
            warn("Ignoring Metadata packet: "+repr(p.data))
            return
        self.metadata_function(p.data[1])


def chunk_message(message):
    """Split an AMF message into 128-byte chunks."""
    chunks = _split_len(message, 128)
    return chunks


def _split_len(seq, length):
    """Split a sequence into len-sized pieces.

    From http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/496784

    """
    return [seq[i:i+length] for i in range(0, len(seq), length)]
